//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  reduce_spec.cpp
//! \brief collection NGeni of generalized integrals

#include "defs.hpp"

#include <boost/format.hpp>
#include <map>

#include "../trivia/string_ops.hpp"

#include "coord.hpp"
#include "obj.hpp"
#include "ptr.hpp"
#include "reduce_spec.hpp"

using boost::format;

//**************************************************************************************************
//*   Low-level implementation
//**************************************************************************************************

//! Compute minimum value of a vector.

RObjNum geni_valmin(int n, const vector<RObjVecNum>& a)
{
    double r = a[0]->to_r(0);
    double dr = a[0]->to_dr(0);
    for (int i = 1; i < n; ++i) {
        if (a[0]->to_r(i) < r) {
            r = a[0]->to_r(i);
            dr = a[0]->to_dr(i);
        }
    }
    return RObjEnu(new CObjEnu(r,dr));
}

//! Compute maximum value of a vector.

RObjNum geni_valmax(int n, const vector<RObjVecNum>& a)
{
    double r = a[0]->to_r(0);
    double dr = a[0]->to_dr(0);
    for (int i = 1; i < n; ++i) {
        if (a[0]->to_r(i) > r) {
            r = a[0]->to_r(i);
            dr = a[0]->to_dr(i);
        }
    }
    return RObjEnu(new CObjEnu(r,dr));
}

//! Return x[i] with i for which y[i] is minimal.

RObjNum geni_argmin(int n, const vector<RObjVecNum>& a)
{
    int idx = 0;
    double extr = a[1]->to_r(0);
    for (int i = 1; i < n; ++i) {
        if (a[1]->to_r(i) < extr) {
            extr = a[1]->to_r(i);
            idx = i;
        }
    }
    return RObjDbl(new CObjDbl(a[0]->to_r(idx)));
}

//! Return x[i] with i for which y[i] is maximal.

RObjNum geni_argmax(int n, const vector<RObjVecNum>& a)
{
    int idx = 0;
    double extr = a[1]->to_r(0);
    for (int i = 1; i < n; ++i) {
        if (a[1]->to_r(i) > extr) {
            extr = a[1]->to_r(i);
            idx = i;
        }
    }
    return RObjDbl(new CObjDbl(a[0]->to_r(idx)));
}

//! Return i for which x[i] is minimal.

RObjNum geni_idxmin(int n, const vector<RObjVecNum>& a)
{
    int idx = 0;
    double extr = a[0]->to_r(0);
    for (int i = 1; i < n; ++i) {
        if (a[0]->to_r(i) < extr) {
            extr = a[0]->to_r(i);
            idx = i;
        }
    }
    return RObjInt(new CObjInt(idx));
}

//! Return i for which x[i] is maximal.

RObjNum geni_idxmax(int n, const vector<RObjVecNum>& a)
{
    int idx = 0;
    double extr = a[0]->to_r(0);
    for (int i = 1; i < n; ++i) {
        if (a[0]->to_r(i) > extr) {
            extr = a[0]->to_r(i);
            idx = i;
        }
    }
    return RObjInt(new CObjInt(idx));
}

//! Sum.

RObjNum geni_sum(int n, const vector<RObjVecNum>& a)
{
    double r = 0;
    for (int i = 0; i < n; i++)
        r += a[0]->to_r(i);
    return RObjNum(new CObjDbl(r));
    // TODO error estimate
}

//! Average.

RObjNum geni_avge(int n, const vector<RObjVecNum>& a)
{
    double r = 0;
    if (!n)
        throw S("Cannot compute average of empty vector");
    for (int i = 0; i < n; i++)
        r += a[0]->to_r(i);
    return RObjNum(new CObjDbl(r/n));
    // TODO error estimate
}

//! Sample variance.

RObjNum geni_variance(int n, const vector<RObjVecNum>& a)
{
    if (n < 2)
        throw S("Cannot compute sample variance for sample size < 2");
    double p = 0, pp = 0, q = 0;
    for (int i = 0; i < n; i++) {
        pp = p + (a[0]->to_r(i) - p) / (i + 1);
        q += (a[0]->to_r(i) - p) * (a[0]->to_r(i) - pp);
        p = pp;
    }
    return RObjDbl(new CObjDbl(q / (n - 1)));
    // TODO error estimate
}

//! Sample standard deviation.

RObjNum geni_stdv(int n, const vector<RObjVecNum>& a)
{
    RObjNum var = geni_variance(n, a);
    return RObjDbl(new CObjDbl(var->to_r()));
    // TODO error estimate
}

//! Integral.

RObjNum geni_integral(int n, const vector<RObjVecNum>& a)
{
    double r = 0;
    for (int i = 0; i < n - 1; i++) {
        if (a[0]->to_r(i + 1) <= a[0]->to_r(i))
            throw S("a not sorted");
        r += (a[0]->to_r(i + 1) - a[0]->to_r(i)) * (a[1]->to_r(i + 1) + a[1]->to_r(i)) / 2;
    }
    return RObjDbl(new CObjDbl(r));
    // TODO error estimate
}

//! Center of gravity in x, weighed by y.

RObjNum geni_cog(int n, const vector<RObjVecNum>& a)
{
    double r0 = 0, r1 = 0;
    for (int i = 0; i < n; i++) {
        r0 += a[1]->to_r(i);
        r1 += a[1]->to_r(i) * a[0]->to_r(i);
    }
    return RObjDbl(new CObjDbl(r0 > 0 ? r1 / r0 : 0));
    // TODO error estimate
}

//! Standard deviation in x, weighed by y.

RObjNum geni_width(int n, const vector<RObjVecNum>& a)
{
    double r0 = 0, r1 = 0, r2 = 0;
    for (int i = 0; i < n; i++) {
        r0 += a[1]->to_r(i);
        r1 += a[1]->to_r(i) * a[0]->to_r(i);
    }
    if (r0 <= 0)
        throw S("r0 <= 0");
    double center = r1 / r0;
    for (int i = 0; i < n; i++) {
        r2 += a[1]->to_r(i) * SQR(a[0]->to_r(i) - center);
    }
    return RObjDbl(new CObjDbl(sqrt(r2 / r0)));
    // TODO error estimate
}

//! Correlation coefficient.
RObjNum geni_corr(int n, const vector<RObjVecNum>& a)
{
    double r0 = 0, r1 = 0, s0 = 0, s1 = 0, v = 0;
    for (int i = 0; i < n; i++) {
        r0 += a[0]->to_r(i) / n;
        r1 += a[1]->to_r(i) / n;
    }
    for (int i = 0; i < n; i++) {
        s0 += SQR(a[0]->to_r(i) - r0) / n;
        s1 += SQR(a[1]->to_r(i) - r1) / n;
        v += (a[0]->to_r(i) - r0) * (a[1]->to_r(i) - r1) / n;
    }
    return RObjDbl(new CObjDbl( v / sqrt(s0 * s1) ));
    // TODO error estimate
}

//! First index for which expr is true.
RObjNum geni_firstwith(int n, const vector<RObjVecNum>& a)
{
    for (int i = 0; i < n; ++i)
        if (a[0]->to_i(i))
            return RObjInt(new CObjInt(i));
    return RObjInt(new CObjInt(-1));
}

//! Last index for which expr is true.
RObjNum geni_lastwith(int n, const vector<RObjVecNum>& a)
{
    for (int i = n-1; i>= 0; --i)
        if (a[0]->to_i(i))
            return RObjInt(new CObjInt(i));
    return RObjInt(new CObjInt(-1));
}


//**************************************************************************************************
//*   Registration and retrieval
//**************************************************************************************************

namespace NGeni
{ // internals:
    std::map<string, const CGeni*> gmap; // unsorted hash, for expression evaluation
    vector<const CGeni*> glist; // sorted array, for help text
}

CGeni::CGeni(
    string _name, int _narg, string _default_arg_str, geni_eval _eval, string _unit_hint,
    string _com)
    : name(_name), narg(_narg), eval(_eval), unit_hint(_unit_hint), com(_com)
{
    default_args = triv::split(_default_arg_str, ",");
    if (default_args.size() != narg)
        throw "BUG: invalid default " + _default_arg_str + " for CGeni " + name;
}

void CGeni::register_me() const { SGeniRegistry::instance()->push_back(name, this); }

void SGeniRegistry::display_entries() const
{
    cout << "Dimension-reducing functionals of data sets.\n";
    cout << "Usage:\n";
    cout << "  - most typically in command 'oi'\n";
    cout << "  - also directly at the command prompt (pocket-calculator mode, only for nj=1)\n";
    cout << "  - or as part of more complex expressions (e.g. 'oy y/integral(x,y)')\n";
    cout << "Arguments:\n";
    cout << "  - most typically just 'x' or 'y',\n";
    cout << "      but any other array is allowed, e.g. 'y[,0]' or 'z0[k+2]'\n";
    cout << "  - list shows default arguments that are assumed when functional name is not\n";
    cout << "      followed by a parenthesis: 'integral' is shorthand for 'integral(x,y)'\n";
    cout << "Implemented functionals:\n";
    for (const CGeni* g : FList) {
        string s1 = g->name + "(" + triv::join(g->default_args, ",") + ")";
        cout << "  " << str(format("%-14s") % s1) << g->com << "\n";
    }
}

void SGeniRegistry::initialize()
{
    // operators by precedence (as in xax_yacc.ypp):
    (new CGeni("valmin", 1, "y", geni_valmin, "a", "minimum value of y"))->register_me();
    (new CGeni("valmax", 1, "y", geni_valmax, "a", "maximum value of y"))->register_me();
    (new CGeni("argmin", 2, "x,y", geni_argmin, "a", "value of x for which y is minimal"))
        ->register_me();
    (new CGeni("argmax", 2, "x,y", geni_argmax, "a", "value of x for which y is maximal"))
        ->register_me();
    (new CGeni("idxmin", 1, "y", geni_idxmin, "-", "index for which x is minimal"))->register_me();
    (new CGeni("idxmax", 1, "y", geni_idxmax, "-", "index for which y is maximal"))->register_me();
    (new CGeni("sum", 1, "y", geni_sum, "a", "sum of y"))->register_me();
    (new CGeni("avge", 1, "y", geni_avge, "a", "average of y"))->register_me();
    (new CGeni("var", 1, "y", geni_variance, "a", "sample variance of y"))->register_me();
    (new CGeni("stdv", 1, "y", geni_stdv, "a", "sample standard deviation of y"))->register_me();
    (new CGeni("integral", 2, "x,y", geni_integral, "ab", "integral d(x) of y"))->register_me();
    (new CGeni("cog", 2, "x,y", geni_cog, "a", "center of gravity of x weighed with y"))
        ->register_me();
    (new CGeni("width", 2, "x,y", geni_width, "a", "standard variation in x weighed with y"))
        ->register_me();
    (new CGeni(
         "corr", 2, "x,y", geni_corr, "?", // TODO: unit?
         "correlation coefficient of x and y"))
        ->register_me();
    (new CGeni("firstwith", 1, "expr", geni_firstwith, "-",
               "first index for which expr is true"))->register_me();
    (new CGeni("lastwith", 1, "expr", geni_firstwith, "-",
               "last index for which expr is true"))->register_me();
}
