//**************************************************************************************************
//* libtrivia: various little routines for C++
//* file_ops: file operations
//* (C) Joachim Wuttke 2001, 2012-
//* http://apps.jcns.fz-juelich.de
//**************************************************************************************************

#include <sys/stat.h>
#include <wordexp.h>
#include <iostream>
#include <boost/format.hpp>
#include <boost/filesystem.hpp>

namespace fs = boost::filesystem;

using boost::format;
using std::string;
using std::vector;
using std::cout;
using std::cerr;

#include "file_ops.hpp"
#include "string_convs.hpp"
#include "string_ops.hpp"


//**************************************************************************************************
//*  improved libc calls
//**************************************************************************************************

//! Execute operation system call, and assert return code 0.

void triv::system(string cmd, bool debug)
{
    if (debug)
        cout << "DEBUG system call '" + cmd + "'\n";
    int ret = ::system(cmd.c_str());
    if (ret)
        throw "Command '" + cmd + "' returned error code " + strg(ret);
    else if (debug)
        cout << "DEBUG system call -> return value " << ret << "\n";
}

//! Execute operation system call, and return its stdout.

string triv::system_read(string cmd, bool debug)
{
    if (debug)
        cout << "DEBUG system call '" + cmd + "'\n";
    FILE* f = ::popen(cmd.c_str(), "r");

    string ret;
    char buf[1024];
    while (fgets(buf, 1024, f))
        ret += buf;

    pclose(f);
    return ret;
}


//**************************************************************************************************
//*  file names, globbing
//**************************************************************************************************

//! Test whether file exists.

bool triv::file_exists(const string& path)
{
    struct stat fileStat;
    if (stat(path.c_str(), &fileStat)) {
        return 0;
    }
    if (!S_ISREG(fileStat.st_mode)) {
        return 0;
    }
    return 1;
}


//! Returns a unique file name obtained by Posix shell expansion of a given pattern.

string triv::wordexp_unique(const string& s)
{
    wordexp_t p;
    int err = wordexp(s.c_str(), &p, 0);
    switch (err) {
    case 0:
        break;
    case WRDE_BADCHAR:
        throw "illegal character in '" + s + "'";
    case WRDE_BADVAL:
        throw "undefined shell variable in '" + s + "'";
    case WRDE_SYNTAX:
        throw "shell syntax error in '" + s + "'";
    default:
        throw "completely unexpected error in '" + s + "'";
    }
    if (p.we_wordc != 1)
        throw "expansion of " + s + " not unique";
    string ret = p.we_wordv[0];
    wordfree(&p);
    return ret;
}


//! Returns a list of file names obtained by Posix shell expansion of a given pattern.

vector<string> triv::wordexp_multi(const string& s)
{
    wordexp_t p;
    if (wordexp(s.c_str(), &p, 0))
        throw "cannot expand " + s;
    if (p.we_wordc == 0)
        throw "expansion of " + s + " is empty";
    vector<string> ret;
    for (size_t i = 0; i < p.we_wordc; ++i)
        if (file_exists(p.we_wordv[i]))
            ret.push_back(p.we_wordv[i]);
    wordfree(&p);
    return ret;
}


//! Construct tmp file name that is not currently in use.

string triv::next_tmp_file(const string& path_format)
{
    const int tmpmax = 1024;
    for (int i = 0; i < tmpmax; ++i) {
        string fname = str(format(path_format) % i);
        if (!triv::file_exists(triv::wordexp_unique(fname)))
            return fname;
        if (i == tmpmax / 2)
            cerr << "WARNING: there more than " << i << " files of form '" << path_format
                 << "'; please clean up soon\n";
    }
    throw "Too many files of type '" + path_format + "' in use";
}


//! Returns a list of file names obtained by shell-like expansion of patterns.

//! Pattern is a blank-separated list of subpatterns.
//! Extensions is a blank-separated list of strings to be appended to subpatterns that have
//! no extension.

vector<string> triv::glob_file_list(const string& patterns, const string& extensions)
{
    string extended_patterns = "";
    vector<string> vPattern = split(patterns);
    if (!vPattern.size())
        throw string("empty file list");
    vector<string> vExtension = split(extensions);
    for (string pat : vPattern) {
        string fext = fs::path(pat).extension().string();
        if (fext == "" && vExtension.size()) {
            for (string ext : vExtension)
                extended_patterns += pat + "." + ext + " ";
        } else {
            extended_patterns += pat + " ";
        }
    }
    return triv::wordexp_multi(extended_patterns);
}


//**************************************************************************************************
//*  File read
//**************************************************************************************************


//! Reads a line from file; returns true unless end-of-file.

bool triv::freadln(FILE* fd, string* s)
{
    char c;
    *s = "";
    while (1) {
        c = (char)fgetc(fd);
        if (c == EOF && *s == "")
            return false;
        else if (c == '\n' || c == EOF)
            return true;
        else if (c == '\r') {
            if (fgetc(fd) == '\n')
                return true;
            else
                throw string("CR not followed by LF");
        }
        *s += c;
    }
}
